@@ -1,6 +1,12 @@
 import logging
 
 import numpy as np
+from scipy.signal import medfilt2d, find_peaks
+from skimage.measure import label
+from skimage.morphology import closing, disk
+
+import multiprocessing as mp
+from itertools import repeat as irepeat
 
 from bmlab import Session
 from bmlab.fits import fit_vipa, VIPA, fit_lorentz_region
@@ -11,24 +17,32 @@ logger = logging.getLogger(__name__)
 
 class ExtractionController(object):
 
+    def __init__(self):
+        self.session = Session.get_instance()
+        return
+
     def add_point(self, calib_key, point):
-        session = Session.get_instance()
-        time = session.get_calibration_time(calib_key)
-        em = session.extraction_model()
+        time = self.session.get_calibration_time(calib_key)
+        if time is None:
+            return
+        em = self.session.extraction_model()
         em.add_point(calib_key, time, *point)
 
     def set_point(self, calib_key, index, point):
-        session = Session.get_instance()
-        time = session.get_calibration_time(calib_key)
-        em = session.extraction_model()
+        time = self.session.get_calibration_time(calib_key)
+        if time is None:
+            return
+        em = self.session.extraction_model()
         em.set_point(calib_key, index, time, *point)
 
     def optimize_points(self, calib_key, radius=10):
-        session = Session.get_instance()
-        em = session.extraction_model()
+        em = self.session.extraction_model()
 
-        imgs = session.get_calibration_image(calib_key)
+        imgs = self.session.get_calibration_image(calib_key)
+        if imgs is None:
+            return
         img = np.nanmean(imgs, axis=0)
+        img = medfilt2d(img)
 
         points = em.get_points(calib_key)
         time = em.get_time(calib_key)
@@ -40,34 +54,176 @@ class ExtractionController(object):
             em.add_point(
                 calib_key, time, new_point[0], new_point[1])
 
+    def find_points_all(self):
+        for calib_key in self.session.get_calib_keys():
+            self.find_points(calib_key)
+
+    def find_points(self, calib_key, min_height=10,
+                    min_area=20, max_distance=50):
+        em = self.session.extraction_model()
+
+        imgs = self.session.get_calibration_image(calib_key)
+        if imgs is None:
+            return
+        img = np.nanmean(imgs, axis=0)
+        time = self.session.get_calibration_time(calib_key)
+        if time is None:
+            return
+
+        img = medfilt2d(img)
+        threshold = np.median(img)
+        img_closed = closing(img > (threshold + min_height), disk(10))
+
+        # Find all peaks higher than the min_height
+        all_peaks = []
+        image_label, num = label(img_closed, return_num=True)
+
+        for region in range(1, num + 1):
+            # Mask of everything but the peak
+            mask = (image_label != region)
+            # Set everything but the peak to zero
+            tmp = img.copy()
+            tmp[mask] = 0
+
+            # Discard all peaks with an area that is too small
+            if np.sum(np.logical_not(mask)) >= min_area:
+                # Find indices of peak maximum
+                ind = np.unravel_index(np.argmax(tmp, axis=None), tmp.shape)
+
+                all_peaks.append(ind)
+
+        # Filter found peaks
+        p0 = (0, img.shape[1])
+        p1 = (img.shape[0], 0)
+        peaks = filter(
+            lambda peak:
+            self.distance_point_to_line(peak, p0, p1) < max_distance,
+            all_peaks)
+
+        # Add found peaks to model
+        em.clear_points(calib_key)
+
+        for p in peaks:
+            # Warning: x-axis in imshow is 1-axis in img, y-axis is 0-axis
+            em.add_point(
+                calib_key, time, p[0], p[1])
+
+    def distance_point_to_line(self, point, line0, line1):
+        return abs(
+            (line1[1] - line0[1]) * (line0[0] - point[0]) -
+            (line0[1] - point[1]) * (line1[0] - line0[0]))\
+               / np.sqrt((line1[1] - line0[1])**2 + (line1[0] - line0[0])**2)
+
 
 class CalibrationController(object):
 
     def __init__(self):
         self.session = Session.get_instance()
-        self.setup = self.session.setup
         return
 
+    def find_peaks(self, calib_key, min_prominence=15,
+                   num_brillouin_samples=2):
+        spectra = self.extract_calibration_spectra(calib_key)
+        if spectra is None:
+            return
+        spectrum = np.mean(spectra, axis=0)
+        peaks, properties = find_peaks(
+            spectrum, prominence=min_prominence, width=True)
+
+        # Number of peaks we are searching for
+        # (2 Rayleigh + 2 times number calibration samples)
+        num_peaks_brillouin = 2 * num_brillouin_samples
+        num_peaks_rayleigh = 2
+        num_peaks = num_peaks_rayleigh + num_peaks_brillouin
+
+        # Check if we found enough peak candidates
+        if len(peaks) < num_peaks:
+            return
+
+        # We need to identify the position between the
+        # Stokes and Anti-Stokes Brillouin peaks
+
+        # In case there are just enough peaks,
+        # we use the position in the middle:
+        if len(peaks) == num_peaks:
+            idx = int(num_peaks / 2)
+            center = np.mean(peaks[idx - 1:idx + 1])
+        # Otherwise we use the center of mass as the middle
+        else:
+            # Subtract background value so it does not affect the center
+            spectrum = spectrum - np.nanmedian(spectrum)
+            # Calculate the center of mass
+            center = np.nansum(spectrum * range(1, len(spectrum) + 1))\
+                / np.nansum(spectrum)
+
+            # Check that we have enough peaks on both sides of the center
+            num_peaks_right = len(peaks[peaks > center])
+            num_peaks_left = len(peaks[peaks <= center])
+
+            # If not enough peaks on the right, shift center to the left
+            if num_peaks_right < (num_brillouin_samples + 1):
+                center = np.mean(
+                    peaks[num_brillouin_samples:num_brillouin_samples+2]
+                )
+            # If not enough peaks on the left, shift center to the right
+            elif num_peaks_left < (num_brillouin_samples + 1):
+                center = np.mean(
+                    peaks[-(num_brillouin_samples+2):-num_brillouin_samples]
+                )
+
+        num_peaks_left = len(peaks[peaks <= center])
+
+        indices_brillouin = range(
+            num_peaks_left - num_brillouin_samples,
+            num_peaks_left + num_brillouin_samples
+        )
+        indices_rayleigh = [
+            num_peaks_left - num_brillouin_samples - 1,
+            num_peaks_left + num_brillouin_samples
+        ]
+
+        def peak_to_region(i):
+            r = (
+                        peaks[i]
+                        + properties['widths'][i] * np.array((-4, 4))
+                ).astype(int)
+            r[r > len(spectrum)] = len(spectrum)
+            return tuple(r)
+
+        regions_brillouin = list(map(peak_to_region, indices_brillouin))
+        # Merge the Brillouin regions if necessary
+        if num_brillouin_samples > 1:
+            regions_brillouin = [
+                (regions_brillouin[0][0],
+                 regions_brillouin[num_brillouin_samples - 1][1]),
+                (regions_brillouin[num_brillouin_samples][0],
+                 regions_brillouin[-1][1]),
+            ]
+
+        regions_rayleigh = map(peak_to_region, indices_rayleigh)
+
+        cm = self.session.calibration_model()
+        # Add Brillouin regions
+        for i, region in enumerate(regions_brillouin):
+            # We use "set_brillouin_region" here so overlapping
+            # regions don't get merged
+            cm.set_brillouin_region(calib_key, i, region)
+        # Add Rayleigh regions
+        for i, region in enumerate(regions_rayleigh):
+            # We use "set_brillouin_region" here so overlapping
+            # regions don't get merged
+            cm.set_rayleigh_region(calib_key, i, region)
+
     def calibrate(self, calib_key, count=None, max_count=None):
 
-        if not calib_key:
-            if max_count is not None:
-                max_count.value = -1
-            return
-
-        if not self.setup:
-            if max_count is not None:
-                max_count.value = -1
-            return
-
+        setup = self.session.setup
         em = self.session.extraction_model()
-        if not em:
-            if max_count is not None:
-                max_count.value = -1
-            return
-
         cm = self.session.calibration_model()
-        if not cm:
+
+        if not calib_key\
+                or not setup\
+                or not em\
+                or not cm:
             if max_count is not None:
                 max_count.value = -1
             return
@@ -75,12 +231,7 @@ class CalibrationController(object):
         spectra = self.extract_calibration_spectra(calib_key)
         time = self.session.get_calibration_time(calib_key)
 
-        if spectra is None:
-            if max_count is not None:
-                max_count.value = -1
-            return
-
-        if len(spectra) == 0:
+        if spectra is None or len(spectra) == 0:
             if max_count is not None:
                 max_count.value = -1
             return
@@ -97,13 +248,13 @@ class CalibrationController(object):
         for frame_num, spectrum in enumerate(spectra):
             peaks = cm.get_sorted_peaks(calib_key, frame_num)
 
-            params = fit_vipa(peaks, self.setup)
+            params = fit_vipa(peaks, setup)
             if params is None:
                 continue
             vipa_params.append(params)
             xdata = np.arange(len(spectrum))
 
-            frequencies.append(VIPA(xdata, params) - self.setup.f0)
+            frequencies.append(VIPA(xdata, params) - setup.f0)
             if count is not None:
                 count.value += 1
 
@@ -141,6 +292,9 @@ class CalibrationController(object):
         cm = self.session.calibration_model()
         spectra = cm.get_spectra(calib_key)
         regions = cm.get_brillouin_regions(calib_key)
+        setup = self.session.setup
+        if not setup:
+            return
 
         cm.clear_brillouin_fits(calib_key)
         for frame_num, spectrum in enumerate(spectra):
@@ -151,7 +305,7 @@ class CalibrationController(object):
                         region,
                         xdata,
                         spectrum,
-                        self.setup.calibration.num_brillouin_samples
+                        setup.calibration.num_brillouin_samples
                     )
                 cm.add_brillouin_fit(calib_key, region_key, frame_num,
                                      w0s, fwhms, intensities, offset)
@@ -183,6 +337,33 @@ class CalibrationController(object):
         return spectra
 
 
+class PeakSelectionController(object):
+
+    def __init__(self):
+        self.session = Session.get_instance()
+        return
+
+    def add_brillouin_region_frequency(self, region_frequency):
+        cm = self.session.calibration_model()
+        # We use the first measurement image here
+        time = self.session.get_payload_time('0')
+        region_pix = tuple(
+            cm.get_position_by_time(time, list(region_frequency)))
+
+        psm = self.session.peak_selection_model()
+        psm.add_brillouin_region(region_pix)
+
+    def add_rayleigh_region_frequency(self, region_frequency):
+        cm = self.session.calibration_model()
+        # We use the first measurement image here
+        time = self.session.get_payload_time('0')
+        region_pix = tuple(
+            cm.get_position_by_time(time, list(region_frequency)))
+
+        psm = self.session.peak_selection_model()
+        psm.add_rayleigh_region(region_pix)
+
+
 class EvaluationController(object):
 
     def __init__(self):
@@ -244,7 +425,11 @@ class EvaluationController(object):
             'nr_rayleigh_regions': len(rayleigh_regions),
         })
 
+        pool_size = mp.cpu_count() * 2
+        pool = mp.Pool(processes=pool_size)
+
         # Loop over all measurement positions
+        ind_total = 0
         for ind_x in range(resolution[0]):
             for ind_y in range(resolution[1]):
                 for ind_z in range(resolution[2]):
@@ -263,38 +448,59 @@ class EvaluationController(object):
                         times
                     evm.results['intensity'][ind_x, ind_y, ind_z, :, 0, 0] =\
                         intensities
-                    # Loop over all frames per measurement position
+
+                    # Pack the data for parallel processing
+                    regions = brillouin_regions + rayleigh_regions
+                    packed_data = zip(irepeat(spectra), regions)
+                    # Process it
+                    results = pool.starmap(self.fit_spectra, packed_data)
+                    # Unpack the results
                     for frame_num, spectrum in enumerate(spectra):
-                        xdata = np.arange(len(spectrum))
-                        # Evaluate all selected regions
                         for region_key, region in enumerate(brillouin_regions):
                             ind = (ind_x, ind_y, ind_z,
                                    frame_num, region_key, 0)
-                            w0, fwhm, intensity, offset = \
-                                fit_lorentz_region(region, xdata, spectrum)
-                            # Save results into arrays
-                            evm.results['brillouin_peak_position'][ind] = w0
-                            evm.results['brillouin_peak_fwhm'][ind] = fwhm
+                            evm.results['brillouin_peak_position'][ind] =\
+                                results[region_key][frame_num][0]
+                            evm.results['brillouin_peak_fwhm'][ind] =\
+                                results[region_key][frame_num][1]
                             evm.results['brillouin_peak_intensity'][ind] =\
-                                intensity
-                        for region_key, region in enumerate(rayleigh_regions):
-                            ind = (ind_x, ind_y, ind_z,
-                                   frame_num, region_key)
-                            w0, fwhm, intensity, offset = \
-                                fit_lorentz_region(region, xdata, spectrum)
-                            # Save results into arrays
-                            evm.results['rayleigh_peak_position'][ind] = w0
-                            evm.results['rayleigh_peak_fwhm'][ind] = fwhm
+                                results[region_key][frame_num][2]
+
+                        for region_key, region \
+                                in enumerate(rayleigh_regions,
+                                             start=len(brillouin_regions)):
+                            ind = (ind_x, ind_y, ind_z, frame_num,
+                                   region_key - len(brillouin_regions))
+                            evm.results['rayleigh_peak_position'][ind] =\
+                                results[region_key][frame_num][0]
+                            evm.results['rayleigh_peak_fwhm'][ind] =\
+                                results[region_key][frame_num][1]
                             evm.results['rayleigh_peak_intensity'][ind] =\
-                                intensity
+                                results[region_key][frame_num][2]
 
                     if count is not None:
                         count.value += 1
 
+                    # Calculate the derived values every ten steps
+                    ind_total += 1
+                    if (ind_total % 10) == 0:
+                        calculate_derived_values()
+
+        pool.close()
+        pool.join()
         calculate_derived_values()
 
         return
 
+    @staticmethod
+    def fit_spectra(spectra, region):
+        fits = []
+        for frame_num, spectrum in enumerate(spectra):
+            xdata = np.arange(len(spectrum))
+            fit = fit_lorentz_region(region, xdata, spectrum)
+            fits.append(fit)
+        return fits
+
     def extract_payload_spectra(self, image_key):
         em = self.session.extraction_model()
         evm = self.session.evaluation_model()
@@ -463,3 +669,52 @@ def calculate_fwhm_f(time, peak_position, peak_fwhm):
             brillouin_peak_left_slope_f
         )
     return np.full(np.shape(peak_position), np.nan)
+
+
+class Controller(object):
+
+    def __init__(self):
+        self.session = Session.get_instance()
+        return
+
+    def evaluate(self, filepath, setup, orientation,
+                 brillouin_regions, rayleigh_regions, repetitions=None):
+        # Load data file
+        self.session.set_file(filepath)
+
+        # Evaluate all repetitions if not requested differently
+        if repetitions is None:
+            repetitions = self.session.file.repetition_keys()
+
+        for repetition in repetitions:
+            # Select repetition
+            self.session.set_current_repetition(repetition)
+            self.session.set_setup(setup)
+
+            # Set orientation
+            self.session.orientation = orientation
+
+            ec = ExtractionController()
+            cc = CalibrationController()
+            psc = PeakSelectionController()
+            evc = EvaluationController()
+
+            # First add all extraction points because this
+            # can influence the extraction for other calibrations
+            ec.find_points_all()
+
+            # Then do the calibration
+            for calib_key in self.session.get_calib_keys():
+                cc.find_peaks(calib_key)
+
+                cc.calibrate(calib_key)
+
+            for region in brillouin_regions:
+                psc.add_brillouin_region_frequency(region)
+
+            for region in rayleigh_regions:
+                psc.add_rayleigh_region_frequency(region)
+
+            evc.evaluate()
+
+        return self.session
