@@ -44,7 +44,10 @@ class CalibrationModel(Serializer):
 
         region = tuple(round(x) for x in region)
 
-        self.brillouin_regions[calib_key][index] = region
+        if index < len(self.brillouin_regions[calib_key]):
+            self.brillouin_regions[calib_key][index] = region
+        else:
+            self.brillouin_regions[calib_key].append(region)
 
     def get_brillouin_regions(self, calib_key):
         regions = self.brillouin_regions.get(calib_key)
@@ -82,7 +85,10 @@ class CalibrationModel(Serializer):
 
         region = tuple(round(x) for x in region)
 
-        self.rayleigh_regions[calib_key][index] = region
+        if index < len(self.rayleigh_regions[calib_key]):
+            self.rayleigh_regions[calib_key][index] = region
+        else:
+            self.rayleigh_regions[calib_key].append(region)
 
     def get_rayleigh_regions(self, calib_key):
         regions = self.rayleigh_regions.get(calib_key)
@@ -181,7 +187,7 @@ class CalibrationModel(Serializer):
         """
         for calib_key in self.frequencies:
             frequencies = self.get_frequencies_by_calib_key(calib_key)
-            if frequencies is None:
+            if frequencies is None or not frequencies:
                 return
             frequency = np.mean(np.array(frequencies), axis=0)
 
@@ -333,6 +339,23 @@ class CalibrationModel(Serializer):
         if self.frequency_by_time_interpolator is not None:
             return self.frequency_by_time_interpolator(time, position)
 
+    def get_position_by_time(self, time, frequencies):
+        """
+        Returns the position on the spectrum in pix for a given frequency
+        at a certain time
+
+        Parameters
+        ----------
+        time: The time
+        frequencies: [GHz] The frequencies we want to convert to pixel
+
+        Returns
+        -------
+
+        """
+        spectrum = self.get_frequencies_by_time(time)
+        return list(map(lambda f: np.argmin(abs(spectrum - f)), frequencies))
+
 
 class FitSet(Serializer):
 
