@@ -68,6 +68,9 @@ class Session(Serializer):
         - a file gets loaded
         - the image orientation changes
         """
+        if self.file is None:
+            return
+
         repetitions = self.file.repetition_keys()
         for repetition in repetitions:
             imgs = self.file.get_repetition(repetition).payload.get_image('0')
@@ -133,27 +136,39 @@ class Session(Serializer):
             Session.get_instance().file = file
 
     def get_calib_keys(self):
+        if self.current_repetition() is None:
+            return None
         return self.current_repetition().calibration.image_keys()
 
     def get_calibration_image(self, calib_key, frame_num=None):
+        if self.current_repetition() is None:
+            return None
         imgs = self.current_repetition().calibration.get_image(calib_key)
         if frame_num is not None:
             imgs = imgs[frame_num, ...]
         return self.orientation.apply(imgs)
 
     def get_calibration_time(self, calib_key):
+        if self.current_repetition() is None:
+            return None
         return self.current_repetition().calibration.get_time(calib_key)
 
     def get_image_keys(self):
+        if self.current_repetition() is None:
+            return None
         return self.current_repetition().payload.image_keys()
 
     def get_payload_image(self, image_key, frame_num=None):
+        if self.current_repetition() is None:
+            return None
         imgs = self.current_repetition().payload.get_image(image_key)
         if frame_num is not None:
             imgs = imgs[frame_num, ...]
         return self.orientation.apply(imgs)
 
     def get_payload_time(self, image_key):
+        if self.current_repetition() is None:
+            return None
         return self.current_repetition().payload.get_time(image_key)
 
     def clear(self):
@@ -194,7 +209,8 @@ class Session(Serializer):
         if self.file is None:
             return
 
-        session_file_name = self.get_session_file_name(self.file.path)
+        session_file_name = self.get_session_file_name(self.file.path,
+                                                       create_folder=True)
 
         with h5py.File(session_file_name, 'w') as f:
             self.serialize(f, 'session', skip=['file'])
@@ -213,13 +229,16 @@ class Session(Serializer):
                 session.__dict__[var_name] = var_value
 
     @staticmethod
-    def get_session_file_name(h5_file_name):
+    def get_session_file_name(h5_file_name, create_folder=False):
         # If the raw data file is located in a 'RawData' folder,
         # we put the eval data file in an 'EvalData' folder and
         # don't append the 'session' string.
         file = Path(h5_file_name)
         if file.parent.name == 'RawData':
-            return str(file.parents[1] /
-                       'EvalData' / (str(file.name)[:-3] + '.h5'))
+            eval_folder = file.parents[1] / 'EvalData'
+            # Create the evaluation folder if necessary
+            if create_folder and not os.path.exists(eval_folder):
+                os.mkdir(eval_folder)
+            return str(eval_folder / (str(file.name)[:-3] + '.h5'))
         else:
             return str(h5_file_name)[:-3] + '.session.h5'
